#!/usr/bin/env python
# -*- coding: utf-8 -*-

# To build simple dialogues etc. (requires pyqt4)
#
#  Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2022 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

haveQt = False  # until we confirm otherwise
importOrder = ['PyQt5', 'PyQt4']

for libname in importOrder:
    try:
        exec("import {}".format(libname))
        haveQt = libname
        break
    except ImportError:
        pass

if not haveQt:
    # do the main import again not in a try...except to recreate error
    exec("import {}".format(importOrder[0]))
elif haveQt == 'PyQt5':
    from PyQt5 import QtWidgets
    from PyQt5 import QtGui
    from PyQt5.QtCore import Qt
else:
    from PyQt4 import QtGui  
    QtWidgets = QtGui  # in qt4 these were all in one package
    from PyQt4.QtCore import Qt

from psychopy import logging
import numpy as np
import os
import sys
import json
from psychopy.localization import _translate

OK = QtWidgets.QDialogButtonBox.Ok

qtapp = QtWidgets.QApplication.instance()


def ensureQtApp():
    global qtapp
    # make sure there's a QApplication prior to showing a gui, e.g., for expInfo
    # dialog
    if qtapp is None:
        qtapp = QtWidgets.QApplication(sys.argv)


wasMouseVisible = True


class Dlg(QtWidgets.QDialog):
    """A simple dialogue box. You can add text or input boxes
    (sequentially) and then retrieve the values.

    see also the function *dlgFromDict* for an **even simpler** version

    **Example**

    .. code-block:: python

        from psychopy import gui

        myDlg = gui.Dlg(title="JWP's experiment")
        myDlg.addText('Subject info')
        myDlg.addField('Name:')
        myDlg.addField('Age:', 21)
        myDlg.addText('Experiment Info')
        myDlg.addField('Grating Ori:',45)
        myDlg.addField('Group:', choices=["Test", "Control"])
        ok_data = myDlg.show()  # show dialog and wait for OK or Cancel
        if myDlg.OK:  # or if ok_data is not None
            print(ok_data)
        else:
            print('user cancelled')

    """

    def __init__(self, title=_translate('PsychoPy Dialog'),
                 pos=None, size=None, style=None,
                 labelButtonOK=_translate(" OK "),
                 labelButtonCancel=_translate(" Cancel "),
                 screen=-1):

        ensureQtApp()
        QtWidgets.QDialog.__init__(self, None, Qt.WindowTitleHint)

        self.inputFields = []
        self.inputFieldTypes = {}
        self.inputFieldNames = []
        self.data = []
        self.irow = 0
        self.pos = pos
        # QtWidgets.QToolTip.setFont(QtGui.QFont('SansSerif', 10))

        # add buttons for OK and Cancel
        self.buttonBox = QtWidgets.QDialogButtonBox(Qt.Horizontal,
                                                    parent=self)
        self.okbutton = QtWidgets.QPushButton(labelButtonOK,
                                              parent=self)
        self.cancelbutton = QtWidgets.QPushButton(labelButtonCancel,
                                                  parent=self)
        self.buttonBox.addButton(self.okbutton,
                                 QtWidgets.QDialogButtonBox.ActionRole)
        self.buttonBox.addButton(self.cancelbutton,
                                 QtWidgets.QDialogButtonBox.ActionRole)
        self.okbutton.clicked.connect(self.accept)
        self.cancelbutton.clicked.connect(self.reject)

        if style:
            raise RuntimeWarning("Dlg does not currently support the "
                                 "style kwarg.")
        self.size = size

        if haveQt == 'PyQt5':
            nScreens = len(qtapp.screens())
        else:
            nScreens = QtWidgets.QDesktopWidget().screenCount()
        self.screen = -1 if screen >= nScreens else screen
        # self.labelButtonOK = labelButtonOK
        # self.labelButtonCancel = labelButtonCancel

        self.layout = QtWidgets.QGridLayout()
        self.layout.setColumnStretch(1, 1)
        self.layout.setSpacing(10)
        self.layout.setColumnMinimumWidth(1, 250)

        self.setLayout(self.layout)

        self.setWindowTitle(title)


    def addText(self, text, color='', isFieldLabel=False):
        textLabel = QtWidgets.QLabel(text, parent=self)

        if len(color):
            palette = QtGui.QPalette()
            palette.setColor(QtGui.QPalette.Foreground, QtGui.QColor(color))
            textLabel.setPalette(palette)

        if isFieldLabel is True:
            self.layout.addWidget(textLabel, self.irow, 0, 1, 1)
        else:
            self.layout.addWidget(textLabel, self.irow, 0, 1, 2)
            self.irow += 1

        return textLabel

    def addField(self, label='', initial='', color='', choices=None, tip='',
                 enabled=True):
        """Adds a (labelled) input field to the dialogue box,
        optional text color and tooltip.

        If 'initial' is a bool, a checkbox will be created.
        If 'choices' is a list or tuple, a dropdown selector is created.
        Otherwise, a text line entry box is created.

        Returns a handle to the field (but not to the label).
        """
        self.inputFieldNames.append(label)
        if choices:
            self.inputFieldTypes[label] = str
        else:
            self.inputFieldTypes[label] = type(initial)
        if type(initial) == np.ndarray:
            initial = initial.tolist()

        # create label
        inputLabel = self.addText(label, color, isFieldLabel=True)

        # create input control
        if type(initial) == bool and not choices:
            self.data.append(initial)
            inputBox = QtWidgets.QCheckBox(parent=self)
            inputBox.setChecked(initial)

            def handleCheckboxChange(new_state):
                ix = self.inputFields.index(inputBox)
                self.data[ix] = inputBox.isChecked()
                msg = "handleCheckboxChange: inputFieldName={0}, checked={1}"
                logging.debug(msg.format(label, self.data[ix]))

            inputBox.stateChanged.connect(handleCheckboxChange)
        elif not choices:
            self.data.append(initial)
            inputBox = QtWidgets.QLineEdit(str(initial), parent=self)

            def handleLineEditChange(new_text):
                ix = self.inputFields.index(inputBox)
                name = self.inputFieldNames[ix]
                thisType = self.inputFieldTypes[name]

                try:
                    if thisType in (str, bytes):
                        self.data[ix] = str(new_text)
                    elif thisType == tuple:
                        jtext = "[" + str(new_text) + "]"
                        self.data[ix] = json.loads(jtext)[0]
                    elif thisType == list:
                        jtext = "[" + str(new_text) + "]"
                        self.data[ix] = json.loads(jtext)[0]
                    elif thisType == float:
                        self.data[ix] = float(new_text)
                    elif thisType == int:
                        self.data[ix] = int(new_text)
                    elif thisType == dict:
                        jtext = "[" + str(new_text) + "]"
                        self.data[ix] = json.loads(jtext)[0]
                    elif thisType == np.ndarray:
                        self.data[ix] = np.array(
                            json.loads("[" + str(new_text) + "]")[0])
                    else:
                        self.data[ix] = new_text
                        msg = ("Unknown type in handleLineEditChange: "
                               "inputFieldName={0}, type={1}, value={2}")
                        logging.warning(msg.format(label, thisType,
                                                   self.data[ix]))
                    msg = ("handleLineEditChange: inputFieldName={0}, "
                           "type={1}, value={2}")
                    logging.debug(msg.format(label, thisType, self.data[ix]))
                except Exception as e:
                    self.data[ix] = str(new_text)
                    msg = ('Error in handleLineEditChange: inputFieldName='
                           '{0}, type={1}, value={2}, error={3}')
                    logging.error(msg.format(label, thisType, self.data[ix],
                                             e))

            inputBox.textEdited.connect(handleLineEditChange)
        else:
            inputBox = QtWidgets.QComboBox(parent=self)
            choices = list(choices)
            for i, option in enumerate(choices):
                inputBox.addItem(str(option))
                # inputBox.addItems([unicode(option) for option in choices])
                inputBox.setItemData(i, (option,))

            if (isinstance(initial, (int, int)) and
                    len(choices) > initial >= 0):
                pass
            elif initial in choices:
                initial = choices.index(initial)
            else:
                initial = 0
            inputBox.setCurrentIndex(initial)

            self.data.append(choices[initial])

            def handleCurrentIndexChanged(new_index):
                ix = self.inputFields.index(inputBox)
                try:
                    self.data[ix] = inputBox.itemData(new_index).toPyObject()[0]
                except AttributeError:
                    self.data[ix] = inputBox.itemData(new_index)[0]
                msg = ("handleCurrentIndexChanged: inputFieldName={0}, "
                       "selected={1}, type: {2}")
                logging.debug(msg.format(label, self.data[ix],
                                         type(self.data[ix])))

            inputBox.currentIndexChanged.connect(handleCurrentIndexChanged)

        if len(color):
            inputBox.setPalette(inputLabel.palette())
        if len(tip):
            inputBox.setToolTip(tip)
        inputBox.setEnabled(enabled)
        self.layout.addWidget(inputBox, self.irow, 1)

        self.inputFields.append(inputBox)  # store this to get data back on OK
        self.irow += 1

        return inputBox

    def addFixedField(self, label='', initial='', color='', choices=None,
                      tip=''):
        """Adds a field to the dialog box (like addField) but the field cannot
        be edited. e.g. Display experiment version.
        """
        return self.addField(label, initial, color, choices, tip,
                             enabled=False)

    def display(self):
        """Presents the dialog and waits for the user to press OK or CANCEL.

        If user presses OK button, function returns a list containing the
        updated values coming from each of the input fields created.
        Otherwise, None is returned.

        :return: self.data
        """
        return self.exec_()

    def show(self):
        """Presents the dialog and waits for the user to press OK or CANCEL.

        If user presses OK button, function returns a list containing the
        updated values coming from each of the input fields created.
        Otherwise, None is returned.

        :return: self.data
        """
        #    NB
        #
        #    ** QDialog already has a show() method. So this method calls
        #       QDialog.show() and then exec_(). This seems to not cause issues
        #       but we need to keep an eye out for any in future.
        return self.display()

    def exec_(self):
        """Presents the dialog and waits for the user to press OK or CANCEL.

        If user presses OK button, function returns a list containing the
        updated values coming from each of the input fields created.
        Otherwise, None is returned.
        """

        self.layout.addWidget(self.buttonBox, self.irow, 0, 1, 2)

        if self.pos is None:
            # Center Dialog on appropriate screen
            frameGm = self.frameGeometry()
            desktop = QtWidgets.QApplication.desktop()
            qtscreen = self.screen
            if self.screen <= 0:
                qtscreen = desktop.primaryScreen()
            centerPoint = desktop.screenGeometry(qtscreen).center()
            frameGm.moveCenter(centerPoint)
            self.move(frameGm.topLeft())
        else:
            self.move(self.pos[0],self.pos[1])
        QtWidgets.QDialog.show(self)
        self.raise_()
        self.activateWindow()

        self.OK = False
        if QtWidgets.QDialog.exec_(self) == QtWidgets.QDialog.Accepted:
            self.OK = True
            return self.data


class DlgFromDict(Dlg):
    """Creates a dialogue box that represents a dictionary of values.
    Any values changed by the user are change (in-place) by this
    dialogue box.

    Parameters
    ----------

    dictionary : dict
        A dictionary defining the input fields (keys) and pre-filled values
        (values) for the user dialog
        
    title : str
        The title of the dialog window

    labels : dict
        A dictionary defining labels (values) to be displayed instead of
        key strings (keys) defined in `dictionary`. Not all keys in
        `dictionary` need to be contained in labels.

    fixed : list
        A list of keys for which the values shall be displayed in non-editable
        fields
    
    order : list
        A list of keys defining the display order of keys in `dictionary`.
        If not all keys in `dictionary`` are contained in `order`, those
        will appear in random order after all ordered keys.

    tip : list
        A dictionary assigning tooltips to the keys

    screen : int
        Screen number where the Dialog is displayed. If -1, the Dialog will
        be displayed on the primary screen.

    sortKeys : bool
        A boolean flag indicating that keys are to be sorted alphabetically.

    copyDict : bool
        If False, modify `dictionary` in-place. If True, a copy of
        the dictionary is created, and the altered version (after
        user interaction) can be retrieved from
        :attr:~`psychopy.gui.DlgFromDict.dictionary`.
        
    labels : dict
        A dictionary defining labels (dict values) to be displayed instead of
        key strings (dict keys) defined in `dictionary`. Not all keys in
        `dictionary´ need to be contained in labels.

    show : bool
        Whether to immediately display the dialog upon instantiation.
        If False, it can be displayed at a later time by calling
        its `show()` method.

    e.g.:

    ::

        info = {'Observer':'jwp', 'GratingOri':45,
                'ExpVersion': 1.1, 'Group': ['Test', 'Control']}
        infoDlg = gui.DlgFromDict(dictionary=info,
                    title='TestExperiment', fixed=['ExpVersion'])
        if infoDlg.OK:
            print(info)
        else:
            print('User Cancelled')

    In the code above, the contents of *info* will be updated to the values
    returned by the dialogue box.

    If the user cancels (rather than pressing OK),
    then the dictionary remains unchanged. If you want to check whether
    the user hit OK, then check whether DlgFromDict.OK equals
    True or False

    See GUI.py for a usage demo, including order and tip (tooltip).
    """

    def __init__(self, dictionary, title='', fixed=None, order=None,
                 tip=None, screen=-1, sortKeys=True, copyDict=False,
                 labels=None, show=True,
                 sort_keys=None, copy_dict=None):

        # We allowed for snake_case parameters in previous releases. This needs
        # to end soon.
        if sort_keys:
            sortKeys = sort_keys
            logging.warning("Parameter 'sort_keys' is deprecated. "
                            "Use 'sortKeys' instead.")

        if copy_dict:
            copyDict = copy_dict
            logging.warning("Parameter 'copy_dict' is deprecated. "
                            "Use 'copyDict' instead.")
        
        # We don't explicitly check for None identity
        # for backward-compatibility reasons.
        if not fixed:
            fixed = []
        if not order:
            order = []
        if not labels:
            labels = dict()
        if not tip:
            tip = dict()

        Dlg.__init__(self, title, screen=screen)

        if copyDict:
            self.dictionary = dictionary.copy()
        else:
            self.dictionary = dictionary

        self._keys = list(self.dictionary.keys())
        self._labels = labels

        if order:
            self._keys = list(order) + list(set(self._keys).difference(set(order)))
        elif sortKeys:
            self._keys.sort()

        for field in self._keys:
            label = labels[field] if field in labels else field
            tooltip = ''
            if field in tip:
                tooltip = tip[field]
            if field in fixed:
                self.addFixedField(label, self.dictionary[field], tip=tooltip)
            elif type(self.dictionary[field]) in [list, tuple]:
                self.addField(label, choices=self.dictionary[field],
                              tip=tooltip)
            else:
                self.addField(label, self.dictionary[field], tip=tooltip)

        if show:
            self.show()

    def show(self):
        """Display the dialog.
        """
        ok_data = self.exec_()
        if ok_data:
            for n, thisKey in enumerate(self._keys):
                if thisKey in self._labels:
                    labelKey = self._labels[thisKey]
                else:
                    labelKey = thisKey
                try:
                    self.dictionary[thisKey] = self.inputFieldTypes[labelKey](self.data[n])
                except ValueError:
                    self.dictionary[thisKey] = self.data[n]


def fileSaveDlg(initFilePath="", initFileName="",
                prompt=_translate("Select file to save"),
                allowed=None):
    """A simple dialogue allowing write access to the file system.
    (Useful in case you collect an hour of data and then try to
    save to a non-existent directory!!)

    :parameters:
        initFilePath: string
            default file path on which to open the dialog
        initFileName: string
            default file name, as suggested file
        prompt: string (default "Select file to open")
            can be set to custom prompts
        allowed: string
            a string to specify file filters.
            e.g. "Text files (\\*.txt) ;; Image files (\\*.bmp \\*.gif)"
            See https://www.riverbankcomputing.com/static/Docs/PyQt4/qfiledialog.html
            #getSaveFileName
            for further details

    If initFilePath or initFileName are empty or invalid then
    current path and empty names are used to start search.

    If user cancels the None is returned.
    """
    if allowed is None:
        allowed = ("All files (*.*);;"
                   "txt (*.txt);;"
                   "pickled files (*.pickle *.pkl);;"
                   "shelved files (*.shelf)")
    ensureQtApp()

    fdir = os.path.join(initFilePath, initFileName)
    pathOut = QtWidgets.QFileDialog.getSaveFileName(parent=None, caption=prompt,
                                              directory=fdir, filter=allowed)
    if type(pathOut) == tuple:  # some versions(?) of PyQt return (files, filter)
        pathOut = pathOut[0]

    if len(pathOut) == 0:
        return None
    return str(pathOut) or None


def fileOpenDlg(tryFilePath="",
                tryFileName="",
                prompt=_translate("Select file to open"),
                allowed=None):
    """A simple dialogue allowing read access to the file system.

    :parameters:

        tryFilePath: string
            default file path on which to open the dialog

        tryFileName: string
            default file name, as suggested file

        prompt: string (default "Select file to open")
            can be set to custom prompts

        allowed: string (available since v1.62.01)
            a string to specify file filters.
            e.g. "Text files (\\*.txt) ;; Image files (\\*.bmp \\*.gif)"
            See https://www.riverbankcomputing.com/static/Docs/PyQt4/qfiledialog.html
            #getOpenFileNames
            for further details

    If tryFilePath or tryFileName are empty or invalid then
    current path and empty names are used to start search.

    If user cancels, then None is returned.
    """
    ensureQtApp()

    if allowed is None:
        allowed = ("All files (*.*);;"
                   "PsychoPy Data (*.psydat);;"
                   "txt (*.txt *.dlm *.csv);;"
                   "pickled files (*.pickle *.pkl);;"
                   "shelved files (*.shelf)")
    fdir = os.path.join(tryFilePath, tryFileName)
    filesToOpen = QtWidgets.QFileDialog.getOpenFileNames(parent=None,
                                                         caption=prompt,
                                                         directory=fdir,
                                                         filter=allowed)
    if type(filesToOpen) == tuple:  # some versions(?) of PyQt return (files, filter)
        filesToOpen = filesToOpen[0]

    filesToOpen = [str(fpath) for fpath in filesToOpen
                   if os.path.exists(fpath)]
    if len(filesToOpen) == 0:
        return None
    return filesToOpen


def infoDlg(title=_translate("Information"), prompt=None):
    ensureQtApp()
    _pr = _translate("No details provided. ('prompt' value not set).")
    QtWidgets.QMessageBox.information(None, title, prompt or _pr)


def warnDlg(title=_translate("Warning"), prompt=None):
    ensureQtApp()
    _pr = _translate("No details provided. ('prompt' value not set).")
    QtWidgets.QMessageBox.warning(None, title, prompt or _pr)


def criticalDlg(title=_translate("Critical"), prompt=None):
    ensureQtApp()
    _pr = _translate("No details provided. ('prompt' value not set).")
    QtWidgets.QMessageBox.critical(None, title, prompt or _pr)


def aboutDlg(title=_translate("About Experiment"), prompt=None):
    ensureQtApp()
    _pr = _translate("No details provided. ('prompt' value not set).")
    QtWidgets.QMessageBox.about(None, title, prompt or _pr)


# Psychopy pyglet window show / hide util functions


def hideWindow(win):
    global wasMouseVisible
    if win.winHandle._visible is True:
        wasMouseVisible = win.mouseVisible
        win.setMouseVisible(True, log=False)
        win.winHandle.minimize()


def showWindow(win):
    global wasMouseVisible
    if win.winHandle._visible is False:
        win.setMouseVisible(wasMouseVisible, log=False)
        win.winHandle.maximize()
        win.winHandle.set_visible(True)
        win.winHandle.activate()


if __name__ == '__main__':
    # Local manual test cases for dialog types....

    logging.console.setLevel(logging.DEBUG)

    # from psychopy import visual, event
    # Create, and hide, a full screen psychopy window
    # win = visual.Window([1024, 768],
    #                    fullscr=True,
    #                    allowGUI=False,
    #                    screen=0)
    # hideWindow(win)

    # Test base Dlg class

    dlg = Dlg()
    dlg.addText("This is a line of text", color="Red")
    dlg.addText("Second line of text")
    dlg.addField("A checkbox", initial=True, tip="Here is your checkbox tip!")
    dlg.addField("Another checkbox", initial=False, color="Blue")
    dlg.addFixedField("ReadOnly checkbox", initial=False, color="Blue",
                      tip="This field is readonly.")
    dlg.addField("A textline", initial="",
                 tip="Here is your <b>textline</b> tip!")
    dlg.addField("A Number:", initial=23, tip="This must be a number.")
    dlg.addField("A List:", initial=[0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                 tip="This must be a list.")
    dlg.addField("A ndarray:",
                 initial=np.asarray([0, 1, 2, 3, 4, 5, 6, 7, 8, 9]),
                 tip="This must be a numpy array.")
    dlg.addField("Another textline", initial="default text", color="Green")
    dlg.addFixedField("ReadOnly textline", initial="default text",
                      tip="This field is readonly.")
    dlg.addField("A dropdown", initial='B', choices=['A', 'B', 'C'],
                 tip="Here is your <b>dropdown</b> tip!")

    dlg.addField("Mixed type dropdown", initial=2,
                 choices=['A String', 1234567, [12.34, 56.78],
                          ('tuple element 0', 'tuple element 1'),
                          {'key1': 'val1', 'key2': 23}],
                 color="Red")

    dlg.addField("Yet Another dropdown", choices=[1, 2, 3])
    dlg.addFixedField("ReadOnly dropdown", initial=2,
                      choices=['R1', 'R2', 'R3'],
                      tip="This field is readonly.")
    ok_data = dlg.show()
    print(("Dlg ok_data:", ok_data))

    # Test Dict Dialog

    info = {'Observer': 'jwp', 'GratingOri': 45,
            'ExpVersion': 1.1, 'Group': ['Test', 'Control']}
    dictDlg = DlgFromDict(dictionary=info, title='TestExperiment',
                          labels={'Group': 'Participant Group'},
                          fixed=['ExpVersion'])
    if dictDlg.OK:
        print(info)
    else:
        print('User Cancelled')

    # Test File Dialogs

    fileToSave = fileSaveDlg(initFileName='__init__.pyc')
    print(("fileToSave: [", fileToSave, "]", type(fileToSave)))

    fileToOpen = fileOpenDlg()
    print(("fileToOpen:", fileToOpen))

    # Test Alert Dialogs

    infoDlg(prompt="Some not important info for you.")

    warnDlg(prompt="Something non critical,\nbut worth mention,\noccurred.")

    _pr = "Oh boy, something really bad just happened:<br><b>{0}</b>"
    criticalDlg(title="RuntimeError",
                prompt=_pr.format(RuntimeError("A made up runtime error")))

    aboutDlg(prompt=u"My Experiment v. 1.0"
             u"<br>"
             u"Written by: Dr. John Doe"
             u"<br>"
             u"Created using <b>PsychoPy</b> © Copyright 2018, Jonathan Peirce")

    # Restore full screen psychopy window
    # showWindow(win)
    # win.flip()
    # from psychopy import event
    # event.waitKeys()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This analysis script takes one or more staircase datafiles as input from a GUI
It then plots the staircases on top of each other on the left
and a combined psychometric function from the same data
on the right.

The combined plot uses every unique X value form the staircase, and alters the
size of the points according to how many trials were run at that level
"""

from psychopy import data, gui, core
from psychopy.tools.filetools import fromFile
import pylab
import os

# set to 0.5 for Yes/No (or PSE). Set to 0.8 for a 2AFC threshold
threshVal = 0.5

# set to zero for Yes/No (or PSE). Set to 0.5 for 2AFC
expectedMin = 0.0

files = gui.fileOpenDlg('.')
if not files:
    core.quit()

# get the data from all the files
allIntensities, allResponses = [], []
for thisFileName in files:
    thisDat = fromFile(thisFileName)
    assert isinstance(thisDat, data.StairHandler)
    allIntensities.append(thisDat.intensities)
    allResponses.append(thisDat.data)
dataFolder = os.path.split(thisFileName)[0]  # just the path, not file name

# plot each staircase in left hand panel
pylab.subplot(121)
colors = 'brgkcmbrgkcm'
lines, names = [], []
for fileN, thisStair in enumerate(allIntensities):
    # lines.extend(pylab.plot(thisStair))  # uncomment for a legend for files
    # names = files[fileN]
    pylab.plot(thisStair, label=files[fileN])
# pylab.legend()

# get combined data
i, r, n = data.functionFromStaircase(allIntensities, allResponses, bins='unique')
combinedInten, combinedResp, combinedN = i, r, n
combinedN = pylab.array(combinedN)  # convert to array so we can do maths

# fit curve
fit = data.FitWeibull(combinedInten, combinedResp, expectedMin=expectedMin,
    sems=1.0 / combinedN)
smoothInt = pylab.arange(min(combinedInten), max(combinedInten), 0.001)
smoothResp = fit.eval(smoothInt)
thresh = fit.inverse(threshVal)
print(thresh)

# plot curve
pylab.subplot(122)
pylab.plot(smoothInt, smoothResp, 'k-')
pylab.plot([thresh, thresh], [0, threshVal], 'k--')  # vertical dashed line
pylab.plot([0, thresh], [threshVal, threshVal], 'k--')  # horizontal dashed line
pylab.title('threshold (%.2f) = %0.3f' %(threshVal, thresh))

# plot points
pointSizes = pylab.array(combinedN) * 5  # 5 pixels per trial at each point
points = pylab.scatter(combinedInten, combinedResp, s=pointSizes,
    edgecolors=(0, 0, 0), facecolor=(1, 1, 1), linewidths=1,
    zorder=10,  # make sure the points plot on top of the line
    )

pylab.ylim([0, 1])
pylab.xlim([0, None])
# save a vector-graphics format for future
outputFile = os.path.join(dataFolder, 'last.pdf')
pylab.savefig(outputFile)
print('saved figure to: ' + outputFile)
pylab.show()

core.quit()

# The contents of this file are in the public domain.



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2022 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).
import io
import sys
import tempfile
import time
import os
import traceback
from pathlib import Path

import gitlab
import requests
from .functions import (setLocalPath, showCommitDialog, logInPavlovia,
                        noGitWarning)
from psychopy.localization import _translate
from psychopy.projects import pavlovia
from psychopy import logging

from psychopy.app.pavlovia_ui import sync, functions

import wx
from wx.lib import scrolledpanel as scrlpanel

from .. import utils
from ...projects.pavlovia import PavloviaProject

try:
    import wx.lib.agw.hyperlink as wxhl  # 4.0+
except ImportError:
    import wx.lib.hyperlink as wxhl  # <3.0.2

_starred = u"\u2605"
_unstarred = u"\u2606"


class ProjectEditor(wx.Dialog):
    def __init__(self, parent=None, id=wx.ID_ANY, project=None, localRoot="",
                 *args, **kwargs):

        wx.Dialog.__init__(self, parent, id,
                           *args, **kwargs)
        panel = wx.Panel(self, wx.ID_ANY, style=wx.TAB_TRAVERSAL)
        # when a project is successfully created these will be populated
        if hasattr(parent, 'filename'):
            self.filename = parent.filename
        else:
            self.filename = None
        self.project = project  # type: pavlovia.PavloviaProject
        self.projInfo = None
        self.parent = parent

        if project:
            # edit existing project
            self.isNew = False
            if project.localRoot and not localRoot:
                localRoot = project.localRoot
        else:
            self.isNew = True

        # create the controls
        nameLabel = wx.StaticText(panel, -1, _translate("Name:"))
        self.nameBox = wx.TextCtrl(panel, -1, size=(400, -1))
        # Path can contain only letters, digits, '_', '-' and '.'.
        # Cannot start with '-', end in '.git' or end in '.atom']
        pavSession = pavlovia.getCurrentSession()

        try:
            username = pavSession.user.username
        except AttributeError as e:
            raise pavlovia.NoUserError("{}: Tried to create project with no user logged in.".format(e))

        gpChoices = [username]
        gpChoices.extend(pavSession.listUserGroups())
        groupLabel = wx.StaticText(panel, -1, _translate("Group/owner:"))
        self.groupBox = wx.Choice(panel, -1, size=(400, -1),
                                  choices=gpChoices)

        descrLabel = wx.StaticText(panel, -1, _translate("Description:"))
        self.descrBox = wx.TextCtrl(panel, -1, size=(400, 200),
                                    style=wx.TE_MULTILINE | wx.SUNKEN_BORDER)

        localLabel = wx.StaticText(panel, -1, _translate("Local folder:"))
        self.localBox = wx.TextCtrl(panel, -1, size=(400, -1),
                                    value=localRoot)
        self.btnLocalBrowse = wx.Button(panel, wx.ID_ANY, _translate("Browse..."))
        self.btnLocalBrowse.Bind(wx.EVT_BUTTON, self.onBrowseLocal)
        localPathSizer = wx.BoxSizer(wx.HORIZONTAL)
        localPathSizer.Add(self.localBox)
        localPathSizer.Add(self.btnLocalBrowse)

        tagsLabel = wx.StaticText(panel, -1,
                                  _translate("Tags (comma separated):"))
        self.tagsBox = wx.TextCtrl(panel, -1, size=(400, 100),
                                   value="PsychoPy, Builder, Coder",
                                   style=wx.TE_MULTILINE | wx.SUNKEN_BORDER)
        publicLabel = wx.StaticText(panel, -1, _translate("Public:"))
        self.publicBox = wx.CheckBox(panel, -1)

        # buttons
        if self.isNew:
            buttonMsg = _translate("Create project on Pavlovia")
        else:
            buttonMsg = _translate("Submit changes to Pavlovia")
        updateBtn = wx.Button(panel, -1, buttonMsg)
        updateBtn.Bind(wx.EVT_BUTTON, self.submitChanges)
        cancelBtn = wx.Button(panel, -1, _translate("Cancel"))
        cancelBtn.Bind(wx.EVT_BUTTON, self.onCancel)
        btnSizer = wx.BoxSizer(wx.HORIZONTAL)
        if sys.platform == "win32":
            btns = [updateBtn, cancelBtn]
        else:
            btns = [cancelBtn, updateBtn]
        btnSizer.AddMany(btns)

        # do layout
        fieldsSizer = wx.FlexGridSizer(cols=2, rows=6, vgap=5, hgap=5)
        fieldsSizer.AddMany([(nameLabel, 0, wx.ALIGN_RIGHT), self.nameBox,
                             (groupLabel, 0, wx.ALIGN_RIGHT), self.groupBox,
                             (localLabel, 0, wx.ALIGN_RIGHT), localPathSizer,
                             (descrLabel, 0, wx.ALIGN_RIGHT), self.descrBox,
                             (tagsLabel, 0, wx.ALIGN_RIGHT), self.tagsBox,
                             (publicLabel, 0, wx.ALIGN_RIGHT), self.publicBox])

        border = wx.BoxSizer(wx.VERTICAL)
        border.Add(fieldsSizer, 0, wx.ALL, 5)
        border.Add(btnSizer, 0, wx.ALIGN_RIGHT | wx.ALL, 5)
        panel.SetSizerAndFit(border)
        self.Fit()

    def onCancel(self, evt=None):
        self.EndModal(wx.ID_CANCEL)

    def submitChanges(self, evt=None):
        session = pavlovia.getCurrentSession()
        if not session.user:
            user = logInPavlovia(parent=self.parent)
        if not session.user:
            return
        # get current values
        name = self.nameBox.GetValue()
        namespace = self.groupBox.GetStringSelection()
        descr = self.descrBox.GetValue()
        visibility = self.publicBox.GetValue()
        # tags need splitting and then
        tagsList = self.tagsBox.GetValue().split(',')
        tags = [thisTag.strip() for thisTag in tagsList]
        localRoot = self.localBox.GetValue()
        if not localRoot:
            localRoot = setLocalPath(self.parent, project=None, path="")

        # then create/update
        if self.isNew:
            project = session.createProject(name=name,
                                            description=descr,
                                            tags=tags,
                                            visibility=visibility,
                                            localRoot=localRoot,
                                            namespace=namespace)
            self.project = project
            self.project._newRemote = True
        else:  # we're changing metadata of an existing project. Don't sync
            self.project.pavlovia.name = name
            self.project.pavlovia.description = descr
            self.project.tags = tags
            self.project.visibility = visibility
            self.project.localRoot = localRoot
            self.project.save()  # pushes changed metadata to gitlab
            self.project._newRemote = False

        self.EndModal(wx.ID_OK)
        pavlovia.knownProjects.save()
        self.project.getRepo(forceRefresh=True)
        self.parent.project = self.project

    def onBrowseLocal(self, evt=None):
        newPath = setLocalPath(self, path=self.filename)
        if newPath:
            self.localBox.SetLabel(newPath)
            self.Layout()
            if self.project:
                self.project.localRoot = newPath
        self.Raise()


class DetailsPanel(wx.Panel):

    class StarBtn(wx.Button):
        def __init__(self, parent, iconCache, value=False):
            wx.Button.__init__(self, parent, label=_translate("Star"))
            # Setup icons
            self.icons = {
                True: iconCache.getBitmap(name="starred", size=16),
                False: iconCache.getBitmap(name="unstarred", size=16),
            }
            self.SetBitmapDisabled(self.icons[False])  # Always appear empty when disabled
            # Set start value
            self.value = value

        @property
        def value(self):
            return self._value

        @value.setter
        def value(self, value):
            # Store value
            self._value = bool(value)
            # Change icon
            self.SetBitmap(self.icons[self._value])
            self.SetBitmapCurrent(self.icons[self._value])
            self.SetBitmapFocus(self.icons[self._value])

        def toggle(self):
            self.value = (not self.value)

    def __init__(self, parent, project=None,
                 size=(650, 550),
                 style=wx.NO_BORDER):

        wx.Panel.__init__(self, parent, -1,
                          size=size,
                          style=style)
        self.SetBackgroundColour("white")
        iconCache = parent.app.iconCache
        # Setup sizer
        self.contentBox = wx.BoxSizer()
        self.SetSizer(self.contentBox)
        self.sizer = wx.BoxSizer(wx.VERTICAL)
        self.contentBox.Add(self.sizer, proportion=1, border=12, flag=wx.ALL | wx.EXPAND)
        # Head sizer
        self.headSizer = wx.BoxSizer(wx.HORIZONTAL)
        self.sizer.Add(self.headSizer, border=0, flag=wx.EXPAND)
        # Icon
        self.icon = utils.ImageCtrl(self, bitmap=wx.Bitmap(), size=(128, 128))
        self.icon.SetBackgroundColour("#f2f2f2")
        self.icon.Bind(wx.EVT_FILEPICKER_CHANGED, self.updateProject)
        self.headSizer.Add(self.icon, border=6, flag=wx.ALL)
        self.icon.SetToolTip(_translate(
            "An image to represent this project, this helps it stand out when browsing on Pavlovia."
        ))
        # Title sizer
        self.titleSizer = wx.BoxSizer(wx.VERTICAL)
        self.headSizer.Add(self.titleSizer, proportion=1, flag=wx.EXPAND)
        # Title
        self.title = wx.TextCtrl(self,
                                 size=(-1, 30 if sys.platform == 'darwin' else -1),
                                 value="")
        self.title.Bind(wx.EVT_KILL_FOCUS, self.updateProject)
        self.title.SetFont(
            wx.Font(24, wx.FONTFAMILY_SWISS, wx.FONTSTYLE_NORMAL, wx.FONTWEIGHT_BOLD)
        )
        self.titleSizer.Add(self.title, border=6, flag=wx.ALL | wx.EXPAND)
        self.title.SetToolTip(_translate(
            "Title of the project. Unlike the project name, this isn't used as a filename anywhere; so you can "
            "add spaces, apostrophes and emojis to your heart's content! 🦕✨"
        ))
        # Author
        self.author = wx.StaticText(self, size=(-1, -1), label="by ---")
        self.titleSizer.Add(self.author, border=6, flag=wx.LEFT | wx.RIGHT)
        # Pavlovia link
        self.link = wxhl.HyperLinkCtrl(self, -1,
                                       label="https://pavlovia.org/",
                                       URL="https://pavlovia.org/",
                                       )
        self.link.SetBackgroundColour("white")
        self.titleSizer.Add(self.link, border=6, flag=wx.EXPAND | wx.LEFT | wx.RIGHT | wx.BOTTOM)
        self.link.SetToolTip(_translate(
            "Click to view the project in Pavlovia."
        ))
        # Button sizer
        self.btnSizer = wx.BoxSizer(wx.HORIZONTAL)
        self.titleSizer.Add(self.btnSizer, flag=wx.EXPAND)
        # Star button
        self.starLbl = wx.StaticText(self, label="-")
        self.btnSizer.Add(self.starLbl, border=6, flag=wx.LEFT | wx.TOP | wx.BOTTOM | wx.ALIGN_CENTER_VERTICAL)
        self.starBtn = self.StarBtn(self, iconCache=iconCache)
        self.starBtn.Bind(wx.EVT_BUTTON, self.star)
        self.btnSizer.Add(self.starBtn, border=6, flag=wx.ALL | wx.EXPAND)
        self.starBtn.SetToolTip(_translate(
            "'Star' this project to get back to it easily. Projects you've starred will appear first in your searches "
            "and projects with more stars in total will appear higher in everyone's searches."
        ))
        # Fork button
        self.forkLbl = wx.StaticText(self, label="-")
        self.btnSizer.Add(self.forkLbl, border=6, flag=wx.LEFT | wx.TOP | wx.BOTTOM | wx.ALIGN_CENTER_VERTICAL)
        self.forkBtn = wx.Button(self, label=_translate("Fork"))
        self.forkBtn.SetBitmap(iconCache.getBitmap(name="fork", size=16))
        self.forkBtn.Bind(wx.EVT_BUTTON, self.fork)
        self.btnSizer.Add(self.forkBtn, border=6, flag=wx.ALL | wx.EXPAND)
        self.forkBtn.SetToolTip(_translate(
            "Create a copy of this project on your own Pavlovia account so that you can make changes without affecting "
            "the original project."
        ))
        # Create button
        self.createBtn = wx.Button(self, label=_translate("Create"))
        self.createBtn.SetBitmap(iconCache.getBitmap(name="plus", size=16))
        self.createBtn.Bind(wx.EVT_BUTTON, self.create)
        self.btnSizer.Add(self.createBtn, border=6, flag=wx.RIGHT | wx.TOP | wx.BOTTOM | wx.ALIGN_CENTER_VERTICAL)
        self.createBtn.SetToolTip(_translate(
            "Create a Pavlovia project for the current experiment."
        ))
        # Sync button
        self.syncBtn = wx.Button(self, label=_translate("Sync"))
        self.syncBtn.SetBitmap(iconCache.getBitmap(name="view-refresh", size=16))
        self.syncBtn.Bind(wx.EVT_BUTTON, self.sync)
        self.btnSizer.Add(self.syncBtn, border=6, flag=wx.ALL | wx.EXPAND)
        self.syncBtn.SetToolTip(_translate(
            "Synchronise this project's local files with their online counterparts. This will 'pull' changes from "
            "Pavlovia and 'push' changes from your local files."
        ))
        # Get button
        self.downloadBtn = wx.Button(self, label=_translate("Download"))
        self.downloadBtn.SetBitmap(iconCache.getBitmap(name="download", size=16))
        self.downloadBtn.Bind(wx.EVT_BUTTON, self.sync)
        self.btnSizer.Add(self.downloadBtn, border=6, flag=wx.ALL | wx.EXPAND)
        self.downloadBtn.SetToolTip(_translate(
            "'Clone' this project, creating local copies of all its files and tracking any changes you make so that "
            "they can be applied when you next 'sync' the project."
        ))
        # Sync label
        self.syncLbl = wx.StaticText(self, size=(-1, -1), label="---")
        self.btnSizer.Add(self.syncLbl, border=6, flag=wx.RIGHT | wx.TOP | wx.BOTTOM | wx.ALIGN_CENTER_VERTICAL)
        self.syncLbl.SetToolTip(_translate(
            "Last synced at..."
        ))
        self.btnSizer.AddStretchSpacer(1)
        # Sep
        self.sizer.Add(wx.StaticLine(self, -1), border=6, flag=wx.EXPAND | wx.ALL)
        # Local root
        self.rootSizer = wx.BoxSizer(wx.HORIZONTAL)
        self.sizer.Add(self.rootSizer, flag=wx.EXPAND)
        self.localRootLabel = wx.StaticText(self, label="Local root:")
        self.rootSizer.Add(self.localRootLabel, border=6, flag=wx.ALIGN_CENTER_VERTICAL | wx.ALL)
        self.localRoot = utils.FileCtrl(self, dlgtype="dir")
        self.localRoot.Bind(wx.EVT_FILEPICKER_CHANGED, self.updateProject)
        self.rootSizer.Add(self.localRoot, proportion=1, border=6, flag=wx.EXPAND | wx.LEFT | wx.RIGHT | wx.BOTTOM)
        self.localRoot.SetToolTip(_translate(
            "Folder in which local files are stored for this project. Changes to files in this folder will be tracked "
            "and applied to the project when you 'sync', so make sure the only files in this folder are relevant!"
        ))
        # Sep
        self.sizer.Add(wx.StaticLine(self, -1), border=6, flag=wx.EXPAND | wx.ALL)
        # Description
        self.description = wx.TextCtrl(self, size=(-1, -1), value="", style=wx.TE_MULTILINE)
        self.description.Bind(wx.EVT_KILL_FOCUS, self.updateProject)
        self.sizer.Add(self.description, proportion=1, border=6, flag=wx.ALL | wx.EXPAND)
        self.description.SetToolTip(_translate(
            "Description of the project to be shown on Pavlovia. Note: This is different than a README file!"
        ))
        # Sep
        self.sizer.Add(wx.StaticLine(self, -1), border=6, flag=wx.EXPAND | wx.ALL)
        # Visibility
        self.visSizer = wx.BoxSizer(wx.HORIZONTAL)
        self.sizer.Add(self.visSizer, flag=wx.EXPAND)
        self.visLbl = wx.StaticText(self, label=_translate("Visibility:"))
        self.visSizer.Add(self.visLbl, border=6, flag=wx.ALIGN_CENTER_VERTICAL | wx.ALL)
        self.visibility = wx.Choice(self, choices=["Private", "Public"])
        self.visibility.Bind(wx.EVT_CHOICE, self.updateProject)
        self.visSizer.Add(self.visibility, proportion=1, border=6, flag=wx.EXPAND | wx.ALL)
        self.visibility.SetToolTip(_translate(
            "Visibility of the current project; whether its visible only to its creator (Private) or to any user "
            "(Public)."
        ))
        # Status
        self.statusSizer = wx.BoxSizer(wx.HORIZONTAL)
        self.sizer.Add(self.statusSizer, flag=wx.EXPAND)
        self.statusLbl = wx.StaticText(self, label=_translate("Status:"))
        self.statusSizer.Add(self.statusLbl, border=6, flag=wx.ALIGN_CENTER_VERTICAL | wx.ALL)
        self.status = wx.Choice(self, choices=["Running", "Piloting", "Inactive"])
        self.status.Bind(wx.EVT_CHOICE, self.updateProject)
        self.statusSizer.Add(self.status, proportion=1, border=6, flag=wx.EXPAND | wx.ALL)
        self.status.SetToolTip(_translate(
            "Project status; whether it can be run to collect data (Running), run by its creator without saving "
            "data (Piloting) or cannot be run (Inactive)."
        ))
        # Tags
        self.tagSizer = wx.BoxSizer(wx.HORIZONTAL)
        self.sizer.Add(self.tagSizer, flag=wx.EXPAND)
        self.tagLbl = wx.StaticText(self, label=_translate("Keywords:"))
        self.tagSizer.Add(self.tagLbl, border=6, flag=wx.ALIGN_CENTER_VERTICAL | wx.ALL)
        self.tags = utils.ButtonArray(self, orient=wx.HORIZONTAL, items=[], itemAlias=_translate("tag"))
        self.tags.Bind(wx.EVT_LIST_INSERT_ITEM, self.updateProject)
        self.tags.Bind(wx.EVT_LIST_DELETE_ITEM, self.updateProject)
        self.tagSizer.Add(self.tags, proportion=1, border=6, flag=wx.EXPAND | wx.ALL)
        self.tags.SetToolTip(_translate(
            "Keywords associated with this project, helping others to find it. For example, if your experiment is "
            "useful to psychophysicists, you may want to add the keyword 'psychophysics'."
        ))
        # Populate
        if project is not None:
            project.refresh()
        self.project = project

    @property
    def project(self):
        return self._project

    @project.setter
    def project(self, project):
        self._project = project

        # Populate fields
        if project is None:
            # Icon
            self.icon.SetBitmap(wx.Bitmap())
            self.icon.SetBackgroundColour("#f2f2f2")
            self.icon.Disable()
            # Title
            self.title.SetValue("")
            self.title.Disable()
            # Author
            self.author.SetLabel("by --- on ---")
            self.author.Disable()
            # Link
            self.link.SetLabel("---/---")
            self.link.SetURL("https://pavlovia.org/")
            self.link.Disable()
            # Star button
            self.starBtn.Disable()
            self.starBtn.value = False
            # Star label
            self.starLbl.SetLabel("-")
            self.starLbl.Disable()
            # Fork button
            self.forkBtn.Disable()
            # Fork label
            self.forkLbl.SetLabel("-")
            self.forkLbl.Disable()
            # Create button
            self.createBtn.Show()
            self.createBtn.Enable(bool(self.session.user))
            # Sync button
            self.syncBtn.Hide()
            # Get button
            self.downloadBtn.Hide()
            # Sync label
            self.syncLbl.SetLabel("---")
            self.syncLbl.Disable()
            # Local root
            self.localRootLabel.Disable()
            wx.TextCtrl.SetValue(self.localRoot, "")  # use base method to avoid callback
            self.localRoot.Disable()
            # Description
            self.description.SetValue("")
            self.description.Disable()
            # Visibility
            self.visibility.SetSelection(wx.NOT_FOUND)
            self.visibility.Disable()
            # Status
            self.status.SetSelection(wx.NOT_FOUND)
            self.status.Disable()
            # Tags
            self.tags.clear()
            self.tags.Disable()
        else:
            # Refresh project to make sure it has info
            if not hasattr(project, "_info"):
                project.refresh()
            # Icon
            if 'avatarUrl' in project.info:
                try:
                    content = requests.get(project['avatar_url']).content
                    icon = wx.Bitmap(wx.Image(io.BytesIO(content)))
                except requests.exceptions.MissingSchema:
                    icon = wx.Bitmap()
            else:
                icon = wx.Bitmap()
            self.icon.SetBitmap(icon)
            self.icon.SetBackgroundColour("#f2f2f2")
            self.icon.Enable(project.editable)
            # Title
            self.title.SetValue(project['name'])
            self.title.Enable(project.editable)
            # Author
            self.author.SetLabel(f"by {project['path_with_namespace'].split('/')[0]} on {project['created_at']:%d %B %Y}")
            self.author.Enable()
            # Link
            self.link.SetLabel(project['path_with_namespace'])
            self.link.SetURL("https://pavlovia.org/" + project['path_with_namespace'])
            self.link.Enable()
            # Star button
            self.starBtn.value = project.starred
            self.starBtn.Enable(bool(project.session.user))
            # Star label
            self.starLbl.SetLabel(str(project['star_count']))
            self.starLbl.Enable()
            # Fork button
            self.forkBtn.Enable(bool(project.session.user) and not project.owned)
            # Fork label
            self.forkLbl.SetLabel(str(project['forks_count']))
            self.forkLbl.Enable()
            # Create button
            self.createBtn.Hide()
            # Sync button
            self.syncBtn.Show(bool(project.localRoot) or (not project.editable))
            self.syncBtn.Enable(project.editable)
            # Get button
            self.downloadBtn.Show(not bool(project.localRoot) and project.editable)
            self.downloadBtn.Enable(project.editable)
            # Sync label
            self.syncLbl.SetLabel(f"{project['last_activity_at']:%d %B %Y, %I:%M%p}")
            self.syncLbl.Show(bool(project.localRoot) or (not project.editable))
            self.syncLbl.Enable(project.editable)
            # Local root
            wx.TextCtrl.SetValue(self.localRoot, project.localRoot or "")  # use base method to avoid callback
            self.localRootLabel.Enable(project.editable)
            self.localRoot.Enable(project.editable)
            # Description
            self.description.SetValue(project['description'])
            self.description.Enable(project.editable)
            # Visibility
            self.visibility.SetStringSelection(project['visibility'])
            self.visibility.Enable(project.editable)
            # Status
            self.status.SetStringSelection(str(project['status2']).title())
            self.status.Enable(project.editable)
            # Tags
            self.tags.items = project['keywords']
            self.tags.Enable(project.editable)

        # Layout
        self.Layout()

    @property
    def session(self):
        # Cache session if not cached
        if not hasattr(self, "_session"):
            self._session = pavlovia.getCurrentSession()
        # Return cached session
        return self._session

    def create(self, evt=None):
        """
        Create a new project
        """
        dlg = sync.CreateDlg(self, user=self.session.user)
        dlg.ShowModal()
        self.project = dlg.project

    def sync(self, evt=None):
        # If not synced locally, choose a folder
        if not self.localRoot.GetValue():
            self.localRoot.browse()
        # If cancelled, return
        if not self.localRoot.GetValue():
            return
        self.project.localRoot = self.localRoot.GetValue()
        # Enable ctrl now that there is a local root
        self.localRoot.Enable()
        self.localRootLabel.Enable()
        # Show sync dlg (does sync)
        dlg = sync.SyncDialog(self, self.project)
        dlg.sync()
        functions.showCommitDialog(self, self.project, initMsg="", infoStream=dlg.status)
        # Update project
        self.project.refresh()
        # Update last sync date & show
        self.syncLbl.SetLabel(f"{self.project['last_activity_at']:%d %B %Y, %I:%M%p}")
        self.syncLbl.Show()
        self.syncLbl.Enable()
        # Switch buttons to show Sync rather than Download/Create
        self.createBtn.Hide()
        self.downloadBtn.Hide()
        self.syncBtn.Show()
        self.syncBtn.Enable()

    def fork(self, evt=None):
        # Do fork
        try:
            proj = self.project.fork()
        except gitlab.GitlabCreateError as e:
            # If project already exists, ask user if they want to view it rather than create again
            dlg = wx.MessageDialog(self, f"{e.error_message}\n\nOpen forked project?", style=wx.YES_NO)
            if dlg.ShowModal() == wx.ID_YES:
                # If yes, show forked project
                projData = requests.get(
                    f"https://pavlovia.org/api/v2/experiments/{self.project.session.user['username']}/{self.project.info['pathWithNamespace'].split('/')[1]}"
                ).json()
                self.project = PavloviaProject(projData['experiment']['gitlabId'])
                return
            else:
                # If no, return
                return
        # Switch to new project
        self.project = proj
        # Sync
        dlg = wx.MessageDialog(self, "Fork created! Sync it to a local folder?", style=wx.YES_NO)
        if dlg.ShowModal() == wx.ID_YES:
            self.sync()

    def star(self, evt=None):
        # Toggle button
        self.starBtn.toggle()
        # Star/unstar project
        self.updateProject(evt)
        # todo: Refresh stars count

    def updateProject(self, evt=None):
        # Skip if no project
        if self.project is None or evt is None:
            return
        # Get object
        obj = evt.GetEventObject()

        # Update project attribute according to supplying object
        if obj == self.title and self.project.editable:
            self.project['name'] = self.title.Value
            self.project.save()
        if obj == self.icon:
            # Create temporary image file
            _, temp = tempfile.mkstemp(suffix=".png")
            self.icon.BitmapFull.SaveFile(temp, wx.BITMAP_TYPE_PNG)
            # Load and upload from temp file
            self.project['avatar'] = open(temp, "rb")
            self.project.save()
            # Delete temp file
            #os.remove(temp)
        if obj == self.starBtn:
            self.project.starred = self.starBtn.value
            self.starLbl.SetLabel(str(self.project.info['nbStars']))
        if obj == self.localRoot:
            if Path(self.localRoot.Value).is_dir():
                self.project.localRoot = self.localRoot.Value
            else:
                dlg = wx.MessageDialog(self,
                                       message=_translate(
                                           "Could not find folder {directory}, please select a different "
                                           "local root.".format(directory=self.localRoot.Value)
                                       ),
                                       caption="Directory not found",
                                       style=wx.ICON_ERROR)
                self.localRoot.SetValue("")
                self.project.localRoot = ""
                dlg.ShowModal()
            # Set project again to trigger a refresh
            self.project = self.project
        if obj == self.description and self.project.editable:
            self.project['description'] = self.description.Value
            self.project.save()
        if obj == self.visibility and self.project.editable:
            self.project['visibility'] = self.visibility.GetStringSelection().lower()
            self.project.save()
        if obj == self.status and self.project.editable:
            requests.put(f"https://pavlovia.org/api/v2/experiments/{self.project.id}",
                         data={"status2": self.status.GetStringSelection()},
                         headers={'OauthToken': self.session.getToken()})
        if obj == self.tags and self.project.editable:
            requests.put(f"https://pavlovia.org/api/v2/experiments/{self.project.id}",
                         data={"keywords": self.tags.GetValue()},
                         headers={'OauthToken': self.session.getToken()})


class ProjectFrame(wx.Dialog):

    def __init__(self, app, parent=None, style=None,
                 pos=wx.DefaultPosition, project=None):
        if style is None:
            style = (wx.DEFAULT_DIALOG_STYLE | wx.CENTER |
                     wx.TAB_TRAVERSAL | wx.RESIZE_BORDER)
        if project:
            title = project['name']
        else:
            title = _translate("Project info")
        self.frameType = 'ProjectInfo'
        wx.Dialog.__init__(self, parent, -1, title=title, style=style,
                           size=(700, 500), pos=pos)
        self.app = app
        self.project = project
        self.parent = parent

        self.detailsPanel = DetailsPanel(parent=self, project=self.project)

        self.mainSizer = wx.BoxSizer(wx.VERTICAL)
        self.mainSizer.Add(self.detailsPanel, proportion=1, border=12, flag=wx.EXPAND | wx.ALL)
        self.SetSizerAndFit(self.mainSizer)

        if self.parent:
            self.CenterOnParent()
        self.Layout()


def syncProject(parent, project, file="", closeFrameWhenDone=False):
    """A function to sync the current project (if there is one)

    Returns
    -----------
        1 for success
        0 for fail
        -1 for cancel at some point in the process
    """
    # If not in a project, make one
    if project is None:
        msgDlg = wx.MessageDialog(parent,
                               message=_translate("This file doesn't belong to any existing project."),
                               style=wx.OK | wx.CANCEL | wx.CENTER)
        msgDlg.SetOKLabel(_translate("Create a project"))
        if msgDlg.ShowModal() == wx.ID_OK:
            # Get start path and name from builder/coder if possible
            if file:
                file = Path(file)
                name = file.stem
                path = file.parent
            else:
                name = path = ""
            # Open dlg to create new project
            createDlg = sync.CreateDlg(parent,
                                       user=pavlovia.getCurrentSession().user,
                                       name=name,
                                       path=path)
            if createDlg.ShowModal() == wx.ID_OK and createDlg.project is not None:
                project = createDlg.project
            else:
                return
        else:
            return
    # If no local root, prompt to make one
    if not project.localRoot:
        defaultRoot = Path(file).parent
        # Ask user if they want to
        dlg = wx.MessageDialog(parent, message=_translate("Project root folder is not yet specified, specify one now?"), style=wx.YES_NO)
        # Open folder picker
        if dlg.ShowModal() == wx.ID_YES:
            dlg = wx.DirDialog(parent, message=_translate("Specify folder..."), defaultPath=str(defaultRoot))
            if dlg.ShowModal() == wx.ID_OK:
                localRoot = Path(dlg.GetPath())
                project.localRoot = str(localRoot)
            else:
                # If cancelled, cancel sync
                return
        else:
            # If they don't want to specify, cancel sync
            return
    # Assign project to parent frame
    parent.project = project
    # If there is (now) a project, do sync
    if project is not None:
        dlg = sync.SyncDialog(parent, project)
        functions.showCommitDialog(parent, project, initMsg="", infoStream=dlg.status)
        dlg.sync()


class ForkDlg(wx.Dialog):
    """Simple dialog to help choose the location/name of a forked project"""
    # this dialog is working fine, but the API call to fork to a specific
    # namespace doesn't appear to work
    def __init__(self, project, *args, **kwargs):
        wx.Dialog.__init__(self, *args, **kwargs)

        existingName = project.name
        session = pavlovia.getCurrentSession()
        groups = [session.user['username']]
        groups.extend(session.listUserGroups())
        msg = wx.StaticText(self, label="Where shall we fork to?")
        groupLbl = wx.StaticText(self, label="Group:")
        self.groupField = wx.Choice(self, choices=groups)
        nameLbl = wx.StaticText(self, label="Project name:")
        self.nameField = wx.TextCtrl(self, value=project.name)

        fieldsSizer = wx.FlexGridSizer(cols=2, rows=2, vgap=5, hgap=5)
        fieldsSizer.AddMany([groupLbl, self.groupField,
                             nameLbl, self.nameField])

        buttonSizer = wx.BoxSizer(wx.HORIZONTAL)
        buttonSizer.Add(wx.Button(self, id=wx.ID_OK, label="OK"))
        buttonSizer.Add(wx.Button(self, id=wx.ID_CANCEL, label="Cancel"))

        mainSizer = wx.BoxSizer(wx.VERTICAL)
        mainSizer.Add(msg, 1, wx.ALL, 5)
        mainSizer.Add(fieldsSizer, 1, wx.ALL, 5)
        mainSizer.Add(buttonSizer, 1, wx.ALL | wx.ALIGN_RIGHT, 5)

        self.SetSizerAndFit(mainSizer)
        self.Layout()


class ProjectRecreator(wx.Dialog):
    """Use this Dlg to handle the case of a missing (deleted?) remote project
    """

    def __init__(self, project, parent, *args, **kwargs):
        wx.Dialog.__init__(self, parent, *args, **kwargs)
        self.parent = parent
        self.project = project
        existingName = project.name
        msgText = _translate("points to a remote that doesn't exist (deleted?).")
        msgText += (" "+_translate("What shall we do?"))
        msg = wx.StaticText(self, label="{} {}".format(existingName, msgText))
        choices = [_translate("(Re)create a project"),
                   "{} ({})".format(_translate("Point to an different location"),
                                    _translate("not yet supported")),
                   _translate("Forget the local git repository (deletes history keeps files)")]
        self.radioCtrl = wx.RadioBox(self, label='RadioBox', choices=choices,
                                     majorDimension=1)
        self.radioCtrl.EnableItem(1, False)
        self.radioCtrl.EnableItem(2, False)

        mainSizer = wx.BoxSizer(wx.VERTICAL)
        buttonSizer = wx.BoxSizer(wx.HORIZONTAL)
        buttonSizer.Add(wx.Button(self, id=wx.ID_OK, label=_translate("OK")),
                      1, wx.ALL, 5)
        buttonSizer.Add(wx.Button(self, id=wx.ID_CANCEL, label=_translate("Cancel")),
                      1, wx.ALL, 5)
        mainSizer.Add(msg, 1, wx.ALL, 5)
        mainSizer.Add(self.radioCtrl, 1, wx.ALL | wx.ALIGN_CENTER_HORIZONTAL, 5)
        mainSizer.Add(buttonSizer, 1, wx.ALL | wx.ALIGN_RIGHT, 1)

        self.SetSizer(mainSizer)
        self.Layout()

    def ShowModal(self):
        if wx.Dialog.ShowModal(self) == wx.ID_OK:
            choice = self.radioCtrl.GetSelection()
            if choice == 0:
                editor = ProjectEditor(parent=self.parent,
                                       localRoot=self.project.localRoot)
                if editor.ShowModal() == wx.ID_OK:
                    self.project = editor.project
                    return 1  # success!
                else:
                    return -1  # user cancelled
            elif choice == 1:
                raise NotImplementedError("We don't yet support redirecting "
                                          "your project to a new location.")
            elif choice == 2:
                raise NotImplementedError("Deleting the local git repo is not "
                                          "yet implemented")
        else:
            return -1



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""This stores variables that are needed globally through the visual package.

It needs to be imported by anything that uses those variables as:

    from psychopy.visual import globalVars
    print(globalVars.currWindow)  # read it
    globalVars.currentWindow = newWin  # change it

Note that if you import using::

    from psychopy.visual.globalVars import currWindow

then if the variable changes your copy won't!
"""




currWindow = None
# for logging purposes, how many times have we resized an image:
nImageResizes = 0




from psychopy import visual
from psychopy.tests import utils
import os
import pytest

v = [(1, 1), (1, -1), (-1, -1), (-1, 1)]  # vertices to use = square
n = 15  # size of the base square
pimg = (n, n)  # position for the image
pgrn = (-n, -n)  # position for green square
img_name = os.path.join(utils.TESTS_DATA_PATH, 'filltext.png')


class Test_Win_Scale_Pos_Ori:
    def setup_class(self):
        self.win = visual.Window(size=(200, 200), units='pix',
                                 allowGUI=False, autoLog=False)

    def teardown_class(self):
        self.win.close()

    @pytest.mark.scalepos
    def test_winScalePosOri(self):
        """test window.viewScale and .viewPos simultaneous
        negative-going scale should mirror-reverse, and position should
        account for that visually, the green square/rect should move clockwise
        around the text

        Non-zero viewOri would not currently pass with a nonzero viewPos
        """
        with pytest.raises(ValueError):
            # units must be define for viewPos!
            _, = visual.Window(size=(200, 200), viewPos=(1, 1), viewOri=1)
        with pytest.raises(NotImplementedError):
            # simultaneous viewPos and viewOri prohibited at present
            _, = visual.Window(size=(200, 200), units='pix',
                               viewPos=(1, 1), viewOri=1)

        for ori in [0, 45]:
            self.win.viewOri = ori
            for offset in [(0, 0), (-0.4, 0)]:
                if ori and (offset[0] or offset[1]):
                    continue  # this combination is NotImplemented
                else:  # NB assumes test win is in pixels!
                    # convert from normalised offset to pixels
                    offset_pix = (round(offs * siz / 2.) for offs, siz in
                                  zip(offset, self.win.size))

                    self.win.viewPos = offset_pix

                for scale in [[1, 1],  # normal: green at lower left
                              [1, -1],  # mirror vert only: green appears to
                                        # move up, text mirrored
                              [-1, -1],  # mirror horiz & vert: green appears
                                         # to move right, text normal but
                                         # upside down
                              [-1, 1],  # mirror horiz only: green appears to
                                        # move down, text mirrored
                              [2, 2],  # same, but both larger
                              [2, -2],
                              [-2, -2],
                              [-2, 2]]:
                    self.win.viewScale = scale
                    self.win.flip()
                    grn = visual.ShapeStim(self.win, vertices=v, pos=pgrn,
                                           size=n, fillColor='darkgreen')
                    img = visual.ImageStim(self.win, image=img_name, size=2*n,
                                           pos=pimg)
                    grn.draw()
                    img.draw()

                    oristr = str(ori)
                    scalestr = str(scale[0]) + '_' + str(scale[1])
                    posstr = str(offset[0]) + '_' + str(offset[1])
                    filename = 'winScalePos_ori%s_scale%s_pos%s.png' % \
                        (oristr, scalestr, posstr)
                    utils.compareScreenshot(filename, self.win, crit=15)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2022 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

from pathlib import Path

from psychopy.experiment import Param
from psychopy.experiment.components import getInitVals, _translate
from psychopy.experiment.components.polygon import PolygonComponent
from psychopy.localization import _localized as __localized
_localized = __localized.copy()

__author__ = 'Jeremy Gray, Jon Peirce'
# March 2011; builder-component for Yuri Spitsyn's visual.Aperture class
# July 2011: jwp added the code for it to be enabled only when needed


class ApertureComponent(PolygonComponent):
    """An event class for using GL stencil to restrict the viewing area to a
    circle or square of a given size and position"""

    categories = ['Stimuli']
    targets = ['PsychoPy']
    iconFile = Path(__file__).parent / 'aperture.png'
    tooltip = _translate('Aperture: restrict the drawing of stimuli to a given '
                         'region')

    def __init__(self, exp, parentName, name='aperture', units='norm',
                 size=1, pos=(0, 0), anchor="center", ori=0,
                 shape='triangle', nVertices=4, vertices="",
                 startType='time (s)', startVal=0.0,
                 stopType='duration (s)', stopVal=1.0,
                 startEstim='', durationEstim=''):
        # initialise main parameters
        super(ApertureComponent, self).__init__(
            exp, parentName, name=name, units=units,
            pos=pos, size=size, ori=ori,
            shape=shape, nVertices=nVertices, vertices=vertices,
            startType=startType, startVal=startVal,
            stopType=stopType, stopVal=stopVal,
            startEstim=startEstim, durationEstim=durationEstim)

        self.type = 'Aperture'
        self.url = "https://www.psychopy.org/builder/components/aperture.html"
        self.order += []

        msg = _translate(
            "How big is the aperture? (a single number for diameter)")
        self.params['size'].hint = msg

        self.params['anchor'] = Param(
            anchor, valType='str', inputType="choice", categ='Layout',
            allowedVals=['center',
                         'top-center',
                         'bottom-center',
                         'center-left',
                         'center-right',
                         'top-left',
                         'top-right',
                         'bottom-left',
                         'bottom-right',
                         ],
            updates='constant',
            hint=_translate("Which point on the aperture should be anchored to its exact position?"),
            label=_translate("Anchor"))

        # only localize hints and labels
        self.params['size'].label = _translate("Size")
        self.params['pos'].hint = _translate("Where is the aperture centred?")

        # Remove Polygon params which are not needed
        del self.params['colorSpace']
        del self.params['fillColor']
        del self.params['lineColor']
        del self.params['lineWidth']
        del self.params['contrast']
        del self.params['opacity']
        del self.params['interpolate']

    def writeInitCode(self, buff):
        # do writing of init
        inits = getInitVals(self.params)

        # additional substitutions
        if self.params['units'].val == 'from exp settings':
            inits['units'].val = None

        if self.params['shape'] == 'regular polygon...':
            inits['vertices'] = self.params['nVertices']
        elif self.params['shape'] != 'custom polygon...':
            inits['vertices'] = self.params['shape']

        code = (
            "%(name)s = visual.Aperture(\n"
        )
        buff.writeIndentedLines(code % inits)

        buff.setIndentLevel(1, relative=True)
        code = (
                "win=win, name='%(name)s',\n"
                "units=%(units)s, size=%(size)s, pos=%(pos)s, ori=%(ori)s,\n"
                "shape=%(vertices)s, anchor=%(anchor)s\n"
        )
        buff.writeIndentedLines(code % inits)

        buff.setIndentLevel(-1, relative=True)
        code = (
                ")\n"
                "%(name)s.disable()  # disable until its actually used\n"
        )
        buff.writeIndentedLines(code % inits)

    def writeFrameCode(self, buff):
        """Only activate the aperture for the required frames
        """
        params = self.params
        code = (f"\n"
                f"# *{params['name']}* updates\n")
        buff.writeIndented(code)
        # writes an if statement to determine whether to draw etc
        self.writeStartTestCode(buff)
        buff.writeIndented("%(name)s.enabled = True\n" % self.params)
        # to get out of the if statement
        buff.setIndentLevel(-1, relative=True)
        if self.params['stopVal'].val not in ['', None, -1, 'None']:
            # writes an if statement to determine whether to draw etc
            self.writeStopTestCode(buff)
            buff.writeIndented("%(name)s.enabled = False\n" % self.params)
            # to get out of the if statement
            buff.setIndentLevel(-2, relative=True)
        # set parameters that need updating every frame
        # do any params need updating? (this method inherited from _base)
        if self.checkNeedToUpdate('set every frame'):
            code = ("if %(name)s.status == STARTED:  # only update if being  drawn\n")
            buff.writeIndented(code % self.params)

            buff.setIndentLevel(+1, relative=True)  # to enter the if block
            self.writeParamUpdates(buff, 'set every frame')
            buff.setIndentLevel(-1, relative=True)  # to exit the if block

    def writeRoutineEndCode(self, buff):
        msg = "%(name)s.enabled = False  # just in case it was left enabled\n"
        buff.writeIndented(msg % self.params)

        # get parent to write code too (e.g. store onset/offset times)
        super().writeRoutineEndCode(buff)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Part of the PsychoPy library
# Copyright (C) 2012-2020 iSolver Software Solutions (C) 2021 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

import numbers  # numbers.Integral is like (int, long) but supports Py3
import os
from collections import namedtuple
import json
import numpy

from ..errors import print2err

from pkg_resources import parse_version
import tables

if parse_version(tables.__version__) < parse_version('3'):
    from tables import openFile as open_file

    walk_groups = "walkGroups"
    list_nodes = "listNodes"
    get_node = "getNode"
    read_where = "readWhere"
else:
    from tables import open_file

    walk_groups = "walk_groups"
    list_nodes = "list_nodes"
    get_node = "get_node"
    read_where = "read_where"

_hubFiles = []


def openHubFile(filepath, filename, mode):
    """
    Open an HDF5 DataStore file and register it so that it is closed even on interpreter crash.
    """
    global _hubFiles
    hubFile = open_file(os.path.join(filepath, filename), mode)
    _hubFiles.append(hubFile)
    return hubFile


def displayDataFileSelectionDialog(starting_dir=None, prompt="Select a ioHub HDF5 File", allowed="HDF5 Files (*.hdf5)"):
    """
    Shows a FileDialog and lets you select a .hdf5 file to open for processing.
    """
    from psychopy.gui.qtgui import fileOpenDlg

    filePath = fileOpenDlg(tryFilePath=starting_dir,
                           prompt=prompt,
                           allowed=allowed)

    if filePath is None:
        return None

    return filePath


def displayEventTableSelectionDialog(title, list_label, list_values, default='Select'):
    from psychopy import gui
    if default not in list_values:
        list_values.insert(0, default)
    else:
        list_values.remove(list_values)
        list_values.insert(0, default)

    selection_dict = {list_label: list_values}
    dlg_info = dict(selection_dict)
    infoDlg = gui.DlgFromDict(dictionary=dlg_info, title=title)
    if not infoDlg.OK:
        return None

    while list(dlg_info.values())[0] == default and infoDlg.OK:
        dlg_info = dict(selection_dict)
        infoDlg = gui.DlgFromDict(dictionary=dlg_info, title=title)

    if not infoDlg.OK:
        return None

    return list(dlg_info.values())[0]


def getEyeSampleTypesInFile(hdf5FilePath):
    """
    Return the eye sample type(s) saved in the hdf5 file located in hdf5FilePath.
    If no eye samples have been saved to the file return []. Possible return list values are defined in
    psychopy.iohub.constants.EYE_SAMPLE_TYPES.

    :param returnType: (type)
    :return: (list)
    """
    dpath, dfile = os.path.split(hdf5FilePath)
    datafile = ExperimentDataAccessUtility(dpath, dfile)
    result = datafile.getAvailableEyeSampleTypes()
    datafile.close()
    return result


def saveEventReport(hdf5FilePath=None, eventType=None, eventFields=[], useConditionsTable=False,
                    usePsychopyDataFile=None, columnNames=[],
                    trialStart=None, trialStop=None, timeMargins=(0.0, 0.0)
                    ):
    """
    Save a tab delimited event report from an iohub .hdf5 data file.


    Events can optionally be split into groups using either a Psychopy .csv data file (usePsychopyDataFile),
    iohub experiment message events, or the hdf5 condition variables table (useConditionsTable=True).

    If usePsychopyDataFile is True, trialStart and trialStop must be provided, or a dialog will prompt the user
    to select a column from the Psychopy .cvs file. The column must have a float or int data type. Each non nan / None
    row will be used to split events.

    If usePsychopyDataFile and useConditionsTable are False and trialStart and trialStop are provided as text,
    events are split based on the time of iohub Experiment Message events that match the trialStart and trialStop text.

    :param hdf5FilePath: (str or None)
    :param eventType: (str or None)
    :param eventFields: (list)
    :param useConditionsTable: (bool)
    :param usePsychopyDataFile: (bool)
    :param columnNames: (list)
    :param trialStart: (str or None)
    :param trialStop: (str or None)
    :param timeMargins: ([float, float] or None)
    :return:
    """
    # Select the hdf5 file to process.
    if usePsychopyDataFile is True and useConditionsTable is True:
        raise RuntimeError("saveEventReport: useConditionsTable and usePsychopyDataFile can both not be True")

    if not hdf5FilePath:
        selectedFilePath = displayDataFileSelectionDialog(os.getcwd())
        if selectedFilePath:
            hdf5FilePath = selectedFilePath[0]
    if not hdf5FilePath:
        raise RuntimeError("Warning: saveEventReport requires hdf5FilePath. No report saved.")

    dpath, dfile = os.path.split(hdf5FilePath)
    datafile = ExperimentDataAccessUtility(dpath, dfile)

    if not eventType:
        # Get a dict of all event types -> DataStore table info for the selected DataStore file.
        eventTableMappings = datafile.getEventMappingInformation()
        # Get event tables that have data...
        events_with_data = datafile.getEventsByType()

        # Select which event table to output
        eventNameList = []
        for event_id in list(events_with_data.keys()):
            eventNameList.append(eventTableMappings[event_id].class_name.decode('utf-8'))
        eventType = displayEventTableSelectionDialog("Select Event Type to Save", "Event Type:", eventNameList)
        if eventType is None:
            datafile.close()
            raise RuntimeError("saveEventReport requires eventType. No report saved.")

    #print("getAvailableEyeSampleTypes: ", datafile.getAvailableEyeSampleTypes())
    # Get the event table to generate report for
    event_table = datafile.getEventTable(eventType)

    if not eventFields:
        # If no event fields were specified, report (almost) all event fields.
        eventFields = [c for c in event_table.colnames if c not in ['experiment_id', 'session_id', 'device_id',
                                                                    'type', 'filter_id']]
    trial_times = []
    column_names = []

    psychoResults = None
    psychopyDataFile = None
    if usePsychopyDataFile is True:
        psychopyDataFile = hdf5FilePath[:-4] + 'csv'
        if not os.path.isfile(psychopyDataFile):
            datafile.close()
            raise RuntimeError("saveEventReport: Could not find .csv file: %s" % psychopyDataFile)

        import pandas
        psychoResults = pandas.read_csv(psychopyDataFile, delimiter=",", encoding='utf-8')

        if trialStart is None or trialStop is None:
            # get list of possible column names from
            columnNames = []
            for columnName in psychoResults.columns:
                if columnName.endswith('.started') or columnName.endswith('.stopped'):
                    if psychoResults[columnName].dtype in [float, int]:
                        columnNames.append(columnName)
            if trialStart is None:
                trialStart = displayEventTableSelectionDialog("Select Event Grouping Start Time Column",
                                                              "Columns", list(columnNames))
            if trialStop is None:
                trialStop = displayEventTableSelectionDialog("Select Event Grouping End Time Column",
                                                              "Columns", [cn for cn in columnNames if cn != trialStart])
                print('trialStop:', trialStop)

        if trialStart and trialStop:
            if trialStart not in psychoResults.columns:
                datafile.close()
                raise ValueError("saveEventReport trialStart column not found in psychopyDataFile: %s" % trialStart)
            if trialStop not in psychoResults.columns:
                datafile.close()
                raise ValueError("saveEventReport trialStop column not found in psychopyDataFile: %s" % trialStop)

            for t_ix, r in psychoResults.iterrows():
                if r[trialStart] != 'None' and r[trialStop] != 'None' and pandas.notna(r[trialStart]) and pandas.notna(
                        r[trialStop]):
                    trial_times.append([t_ix, r[trialStart] - timeMargins[0], r[trialStop] + timeMargins[1]])

        else:
            datafile.close()
            raise ValueError("saveEventReport trialStart and trialStop must be specified when using psychopyDataFile.")

    cvTable = None
    if useConditionsTable is True:
        # Use hdf5 conditions table columns 'trialStart' and 'trialStop' to group events
        if trialStart is None or trialStop is None:
            # If either trialStart and trialStop are None, display selection dialogs
            try:
                cvColumnNames = datafile.getConditionVariableNames()[2:]
            except Exception as e:
                #datastore Conditions table must not exist
                datafile.close()
                raise RuntimeError("saveEventReport: Error calling datafile.getConditionVariableNames().\n{}".format(e))

            if trialStart is None:
                trialStart = displayEventTableSelectionDialog("Select Event Grouping Start Time Column",
                                                              "Columns", list(cvColumnNames))
            if trialStop is None:
                trialStop = displayEventTableSelectionDialog("Select Event Grouping End Time Column",
                                                              "Columns", [cn for cn in cvColumnNames if cn != trialStart])


        if trialStart is None or trialStop is None:
            datafile.close()
            raise ValueError("saveEventReport: trialStart and trialStop must be specified "
                             "when useConditionsTable is True.")

        if trialStart not in cvColumnNames:
            datafile.close()
            raise ValueError("saveEventReport:"
                             " trialStart column not found in trial condition variables table: %s" % trialStart)
        if trialStop not in cvColumnNames:
            datafile.close()
            raise ValueError("saveEventReport:"
                             " trialStop column not found in trial condition variables table: %s" % trialStop)

        cvTable = datafile.getConditionVariablesTable()
        for t_ix, r in enumerate(cvTable):
            trial_times.append([t_ix, r[trialStart] - timeMargins[0], r[trialStop] + timeMargins[1]])

    if useConditionsTable is False and psychoResults is None and trialStart and trialStop:
        # Create a table of trial_index, trial_start_time, trial_end_time for each trial by
        # getting the time of 'TRIAL_START' and 'TRIAL_END' experiment messages.
        mgs_table = datafile.getEventTable('MessageEvent')
        trial_start_msgs = mgs_table.where('text == b"%s"' % trialStart)
        for mix, msg in enumerate(trial_start_msgs):
            trial_times.append([mix + 1, msg['time'] - timeMargins[0], 0])
        trial_end_msgs = mgs_table.where('text == b"%s"' % trialStop)
        for mix, msg in enumerate(trial_end_msgs):
            trial_times[mix][2] = msg['time'] + timeMargins[1]
        del mgs_table
    elif trialStart is None and trialStop is None:
        # do not split events into trial groupings
        pass
    elif trialStart is None or trialStop is None:
        datafile.close()
        raise RuntimeError("Warning: saveEventReport requires trialStart and trialStop to be strings or both None."
                           " No report saved.")

    if eventType == 'MessageEvent':
        # Sort experiment messages by time since they may not be ordered chronologically.
        event_table = event_table.read()
        event_table.sort(order='time')

    ecount = 0
    # Open a file to save the tab delimited output to.
    output_file_name = os.path.join(dpath, "%s.%s.txt" % (dfile[:-5], eventType))
    with open(output_file_name, 'w') as output_file:
        # Save header row to file
        if trial_times:
            if useConditionsTable:
                cvtColumnNames = datafile.getConditionVariableNames()[2:]
                if columnNames:
                    for cname in columnNames:
                        if cname not in cvtColumnNames:
                            datafile.close()
                            raise ValueError("saveEventReport: .hdf5 conditions table column '%s' not found." % cname)
                    column_names = list(columnNames) + eventFields
                else:
                    column_names = list(cvtColumnNames) + eventFields
                    columnNames = list(cvtColumnNames)

            elif hasattr(psychoResults, 'columns'):
                if columnNames:
                    for cname in columnNames:
                        if cname not in psychoResults.columns:
                            datafile.close()
                            raise ValueError(
                                "saveEventReport: psychopyDataFileColumn '%s' not found in .csv file." % cname)
                    column_names = list(columnNames) + eventFields
                else:
                    column_names = list(psychoResults.columns) + eventFields
                    columnNames = list(psychoResults.columns)
            else:
                column_names = ['TRIAL_INDEX', trialStart, trialStop] + eventFields
        else:
            column_names = eventFields

        output_file.write('\t'.join(column_names))
        output_file.write('\n')

        event_groupings = []
        if trial_times:
            # Split events into trials
            for tindex, tstart, tstop in trial_times:
                if eventType == 'MessageEvent':
                    event_groupings.append(event_table[(event_table['time'] >= tstart) & (event_table['time']
                                                                                          <= tstop)])
                else:
                    event_groupings.append(event_table.where("(time >= %f) & (time <= %f)" % (tstart, tstop)))
        else:
            # Report events without splitting them into trials
            if eventType == 'MessageEvent':
                event_groupings.append(event_table)
            else:
                event_groupings.append(event_table.iterrows())

        # Save a row for each event within the trial period
        for tid, trial_events in enumerate(event_groupings):
            for event in trial_events:
                event_data = []
                for c in eventFields:
                    cv = event[c]
                    if type(cv) == numpy.bytes_:
                        cv = event[c].decode('utf-8')
                    if type(cv) == str and len(cv) == 0:
                        cv = '.'
                    event_data.append(str(cv))
                if trial_times:
                    tindex, tstart, tstop = trial_times[tid]
                    if useConditionsTable:
                        cvRow=cvTable.read(tindex, tindex+1)
                        cvrowdat = [cvRow[c][0] for c in columnNames]
                        for ri, cv in enumerate(cvrowdat):
                            if type(cv) == numpy.bytes_:
                                cvrowdat[ri] = cvrowdat[ri].decode('utf-8')
                            else:
                                cvrowdat[ri] = str(cvrowdat[ri])
                            if type(cv) == str and len(cv) == 0:
                                cvrowdat[ri] = '.'
                        output_file.write('\t'.join(cvrowdat + event_data))
                    elif hasattr(psychoResults, 'columns'):
                        drow = psychoResults.iloc[tindex]
                        prowdat = [str(drow[c]) for c in columnNames]
                        output_file.write('\t'.join(prowdat + event_data))
                    else:
                        output_file.write('\t'.join([str(tindex), str(tstart), str(tstop)] + event_data))
                else:
                    output_file.write('\t'.join(event_data))
                output_file.write('\n')
                ecount += 1

    # Done report creation, close input file
    datafile.close()
    return output_file_name, ecount


########### Experiment / Experiment Session Based Data Access #################


class ExperimentDataAccessUtility:
    """The ExperimentDataAccessUtility  provides a simple, high level, way to
    access data saved in an ioHub DataStore HDF5 file. Data access is done by
    providing information at an experiment and session level, as well as
    specifying the ioHub Event types you want to retrieve data for.

    An instance of the ExperimentDataAccessUtility class is created by providing
    the location and name of the file to read, as well as any session code
    filtering you want applied to the retrieved datasets.

    Args:
        hdfFilePath (str): The path of the directory the DataStore HDF5 file is in.

        hdfFileName (str): The name of the DataStore HDF5 file.

        experimentCode (str): If multi-experiment support is enabled for the DataStore file, this argument can be used to specify what experiment data to load based on the experiment_code given. NOTE: Multi-experiment data file support is not well tested and should not be used at this point.

        sessionCodes (str or list): The experiment session code to filter data by. If a list of codes is given, then all codes in the list will be used.

    Returns:
        object: the created instance of the ExperimentDataAccessUtility, ready to get your data!

    """

    def __init__(self, hdfFilePath, hdfFileName, experimentCode=None, sessionCodes=[], mode='r'):
        """An instance of the ExperimentDataAccessUtility class is created by
        providing the location and name of the file to read, as well as any
        session code filtering you want applied to the retrieved datasets.

        Args:
            hdfFilePath (str): The path of the directory the DataStore HDF5 file is in.

            hdfFileName (str): The name of the DataStore HDF5 file.

            experimentCode (str): If multi-experiment support is enabled for the DataStore file, this argument can be used to specify what experiment data to load based on the experiment_code given. NOTE: Multi-experiment data file support is not well tested and should not be used at this point.

            sessionCodes (str or list): The experiment session code to filter data by. If a list of codes is given, then all codes in the list will be used.

        Returns:
            object: the created instance of the ExperimentDataAccessUtility, ready to get your data!

        """
        self.hdfFilePath = hdfFilePath
        self.hdfFileName = hdfFileName
        self.mode = mode
        self.hdfFile = None

        self._experimentCode = experimentCode
        self._sessionCodes = sessionCodes
        self._lastWhereClause = None

        try:
            self.hdfFile = openHubFile(hdfFilePath, hdfFileName, mode)
        except Exception as e:
            raise ExperimentDataAccessException(e)

        self.getExperimentMetaData()

    def printTableStructure(self, tableName):
        """Print to stdout the current structure and content statistics of the
        specified DataStore table. To print out the complete structure of the
        DataStore file, including the name of all available tables, see the
        printHubFileStructure method.

        Args:
            tableName (str): The DataStore table name to print metadata information out for.

        """
        if self.hdfFile:
            hubFile = self.hdfFile
            for group in getattr(hubFile, walk_groups)("/"):
                for table in getattr(hubFile, list_nodes)(group, classname='Table'):
                    if table.name == tableName:
                        print('------------------')
                        print('Path:', table)
                        print('Table name:', table.name)
                        print('Number of rows in table:', table.nrows)
                        print('Number of cols in table:', len(table.colnames))
                        print('Attribute name := type, shape:')
                        for name in table.colnames:
                            print('\t', name, ':= %s, %s' % (table.coldtypes[name], table.coldtypes[name].shape))
                        print('------------------')
                        return

    def printHubFileStructure(self):
        """Print to stdout the current global structure of the loaded DataStore
        File."""
        if self.hdfFile:
            print(self.hdfFile)

    def getExperimentMetaData(self):
        """Returns the the metadata for the experiment the datStore file is
        for.

        **Docstr TBC.**

        """
        if self.hdfFile:
            expcols = self.hdfFile.root.data_collection.experiment_meta_data.colnames
            if 'sessions' not in expcols:
                expcols.append('sessions')
            ExperimentMetaDataInstance = namedtuple(
                'ExperimentMetaDataInstance', expcols)
            experiments = []
            for e in self.hdfFile.root.data_collection.experiment_meta_data:
                self._experimentID = e['experiment_id']
                a_exp = list(e[:])
                a_exp.append(self.getSessionMetaData())
                experiments.append(ExperimentMetaDataInstance(*a_exp))
            return experiments

    def getSessionMetaData(self, sessions=None):
        """
        Returns the the metadata associated with the experiment session codes in use.

        **Docstr TBC.**

        """
        if self.hdfFile:
            if sessions is None:
                sessions = []

            sessionCodes = self._sessionCodes
            sesscols = self.hdfFile.root.data_collection.session_meta_data.colnames
            SessionMetaDataInstance = namedtuple('SessionMetaDataInstance', sesscols)
            for r in self.hdfFile.root.data_collection.session_meta_data:
                if (len(sessionCodes) == 0 or r['code'] in sessionCodes) and r['experiment_id'] == self._experimentID:
                    rcpy = list(r[:])
                    rcpy[-1] = json.loads(rcpy[-1])
                    sessions.append(SessionMetaDataInstance(*rcpy))
            return sessions

    def getTableForPath(self, path):
        """
        Given a valid table path within the DataStore file, return the accociated table.
        """
        getattr(self.hdfFile, get_node)(path)

    def getEventTable(self, event_type):
        """
        Returns the DataStore table that contains events of the specified type.

        **Docstr TBC.**

        """
        if self.hdfFile:
            klassTables = self.hdfFile.root.class_table_mapping
            event_column = None
            event_value = None

            if isinstance(event_type, str):
                if event_type.find('Event') >= 0:
                    event_column = 'class_name'
                    event_value = event_type
                else:
                    event_value = ''
                    tokens = event_type.split('_')
                    for t in tokens:
                        event_value += t[0].upper() + t[1:].lower()
                    event_value = event_type + 'Event'
            elif isinstance(event_type, numbers.Integral):
                event_column = 'class_id'
                event_value = event_type
            else:
                print2err(
                    'getEventTable error: event_type argument must be a string or and int')
                return None

            result = []
            where_cls = '(%s == b"%s") & (class_type_id == 1)' % (event_column, event_value)
            for row in klassTables.where(where_cls):
                result.append(row.fetch_all_fields())

            if len(result) == 0:
                return None

            if len(result) != 1:
                print2err('event_type_id passed to getEventAttribute can only return one row from CLASS_MAPPINGS.')
                return None
            tablePathString = result[0][3]
            if isinstance(tablePathString, bytes):
                tablePathString = tablePathString.decode('utf-8')
            return getattr(self.hdfFile, get_node)(tablePathString)
        return None

    def getEventMappingInformation(self):
        """Returns details on how ioHub Event Types are mapped to tables within
        the given DataStore file."""
        if self.hdfFile:
            eventMappings = dict()
            class_2_table = self.hdfFile.root.class_table_mapping
            EventTableMapping = namedtuple(
                'EventTableMapping',
                self.hdfFile.root.class_table_mapping.colnames)
            for row in class_2_table[:]:
                eventMappings[row['class_id']] = EventTableMapping(*row)
            return eventMappings
        return None

    def getEventsByType(self, condition_str=None):
        """Returns a dict of all event tables within the DataStore file that
        have at least one event instance saved.

        Keys are Event Type constants, as specified by
        iohub.EventConstants. Each value is a row iterator for events of
        that type.

        """
        eventTableMappings = self.getEventMappingInformation()
        if eventTableMappings:
            events_by_type = dict()
            getNode = getattr(self.hdfFile, get_node)
            for event_type_id, event_mapping_info in eventTableMappings.items():
                try:
                    cond = '(type == %d)' % (event_type_id)
                    if condition_str:
                        cond += ' & ' + condition_str
                    et_path = event_mapping_info.table_path
                    if isinstance(et_path, bytes):
                        et_path = et_path.decode('utf-8')
                    events_by_type[event_type_id] = next(getNode(et_path).where(cond))
                except StopIteration:
                    pass
            return events_by_type
        return None

    def getAvailableEyeSampleTypes(self, returnType=str):
        """
        Return the eye sample type(s) saved to the current hdf5 file.
        If no eye samples have been saved to the file return []. Possible return list values are defined in
        psychopy.iohub.constants.EYE_SAMPLE_TYPES.

        :param returnType: (type)
        :return: (list)
        """
        from psychopy.iohub.constants import EYE_SAMPLE_TYPES

        if returnType == int:
            return [etype for etype in self.getEventsByType() if etype in EYE_SAMPLE_TYPES]

        if returnType == str:
            eventTableMappings = self.getEventMappingInformation()
            sampleTypes = [etype for etype in self.getEventsByType() if etype in EYE_SAMPLE_TYPES]
            eventList = []
            for event_id in sampleTypes:
                eventList.append(eventTableMappings[event_id].class_name.decode('utf-8'))
            return eventList

        raise RuntimeError("getAvailableEyeSampleTypes returnType arg must be set to either int or str type.")

    def getConditionVariablesTable(self):
        """
        **Docstr TBC.**
        """
        cv_group = self.hdfFile.root.data_collection.condition_variables
        ecv = 'EXP_CV_%d' % (self._experimentID,)
        if ecv in cv_group._v_leaves:
            return cv_group._v_leaves[ecv]
        return None

    def getConditionVariableNames(self):
        """
        **Docstr TBC.**
        """
        cv_group = self.hdfFile.root.data_collection.condition_variables
        ecv = "EXP_CV_%d" % (self._experimentID,)
        if ecv in cv_group._v_leaves:
            ecvTable = cv_group._v_leaves[ecv]
            return ecvTable.colnames
        return None

    def getConditionVariables(self, filter=None):
        """
        **Docstr TBC.**
        """
        if filter is None:
            session_ids = []
            for s in self.getExperimentMetaData()[0].sessions:
                session_ids.append(s.session_id)
            filter = dict(SESSION_ID=(' in ', session_ids))

        ConditionSetInstance = None

        for conditionVarName, conditionVarComparitor in filter.items():
            avComparison, value = conditionVarComparitor

            cv_group = self.hdfFile.root.data_collection.condition_variables
            cvrows = []
            ecv = "EXP_CV_%d" % (self._experimentID,)
            if ecv in cv_group._v_leaves:
                ecvTable = cv_group._v_leaves[ecv]

                if ConditionSetInstance is None:
                    colnam = ecvTable.colnames
                    ConditionSetInstance = namedtuple('ConditionSetInstance', colnam)

                cvrows.extend(
                    [
                        ConditionSetInstance(
                            *
                            r[:]) for r in ecvTable if all(
                        [
                            eval(
                                '{0} {1} {2}'.format(
                                    r[conditionVarName],
                                    conditionVarComparitor[0],
                                    conditionVarComparitor[1])) for conditionVarName,
                                                                    conditionVarComparitor in filter.items()])])
        return cvrows

    def getValuesForVariables(self, cv, value, cvNames):
        """
        **Docstr TBC.**
        """
        if isinstance(value, (list, tuple)):
            resolvedValues = []
            for v in value:
                if isinstance(value, str) and value.startswith('@') and value.endswith('@'):
                    value = value[1:-1]
                    if value in cvNames:
                        resolvedValues.append(getattr(cv, v))
                    else:
                        raise ExperimentDataAccessException('getEventAttributeValues: {0} is not a valid attribute '
                                                            'name in {1}'.format(v, cvNames))
                elif isinstance(value, str):
                    resolvedValues.append(value)
            return resolvedValues
        elif isinstance(value, str) and value.startswith('@') and value.endswith('@'):
            value = value[1:-1]
            if value in cvNames:
                return getattr(cv, value)
            else:
                raise ExperimentDataAccessException('getEventAttributeValues: {0} is not a valid attribute name'
                                                    ' in {1}'.format(value, cvNames))
        else:
            raise ExperimentDataAccessException('Unhandled value type !: {0} is not a valid type for value '
                                                '{1}'.format(type(value), value))

    def getEventAttributeValues(self, event_type_id, event_attribute_names, filter_id=None,
                                conditionVariablesFilter=None, startConditions=None, endConditions=None):
        """
        **Docstr TBC.**

        Args:
            event_type_id
            event_attribute_names
            filter_id
            conditionVariablesFilter
            startConditions
            endConditions

        Returns:
            Values for the specified event type and event attribute columns which match the provided experiment
            condition variable filter, starting condition filer, and ending condition filter criteria.
        """
        if self.hdfFile:
            klassTables = self.hdfFile.root.class_table_mapping

            deviceEventTable = None

            result = [row.fetch_all_fields() for row in klassTables.where('(class_id == %d) &'
                                                                          ' (class_type_id == 1)' % (event_type_id))]
            if len(result) != 1:
                raise ExperimentDataAccessException("event_type_id returned > 1 row from CLASS_MAPPINGS.")
            tablePathString = result[0][3]
            if isinstance(tablePathString, bytes):
                tablePathString = tablePathString.decode('utf-8')
            deviceEventTable = getattr(self.hdfFile, get_node)(tablePathString)

            for ename in event_attribute_names:
                if ename not in deviceEventTable.colnames:
                    raise ExperimentDataAccessException('getEventAttribute: %s does not have a column named %s' %
                                                        (deviceEventTable.title, event_attribute_names))

            resultSetList = []

            csier = list(event_attribute_names)
            csier.append('query_string')
            csier.append('condition_set')
            EventAttributeResults = namedtuple('EventAttributeResults', csier)

            if deviceEventTable is not None:
                if not isinstance(event_attribute_names, (list, tuple)):
                    event_attribute_names = [event_attribute_names, ]

                filteredConditionVariableList = None
                if conditionVariablesFilter is None:
                    filteredConditionVariableList = self.getConditionVariables()
                else:
                    filteredConditionVariableList = self.getConditionVariables(conditionVariablesFilter)

                cvNames = self.getConditionVariableNames()

                # no further where clause building needed; get reseults and
                # return
                if startConditions is None and endConditions is None:
                    for cv in filteredConditionVariableList:

                        wclause = '( experiment_id == {0} ) & ( SESSION_ID == {1} )'.format(self._experimentID,
                                                                                            cv.SESSION_ID)

                        wclause += ' & ( type == {0} ) '.format(event_type_id)

                        if filter_id is not None:
                            wclause += '& ( filter_id == {0} ) '.format(filter_id)

                        resultSetList.append([])

                        for ename in event_attribute_names:
                            resultSetList[-1].append(getattr(deviceEventTable, read_where)(wclause, field=ename))
                        resultSetList[-1].append(wclause)
                        resultSetList[-1].append(cv)

                        eventAttributeResults = EventAttributeResults(*resultSetList[-1])
                        resultSetList[-1] = eventAttributeResults

                    return resultSetList

                # start or end conditions exist....
                for cv in filteredConditionVariableList:
                    resultSetList.append([])

                    wclause = '( experiment_id == {0} ) & ( session_id == {1} )'.format(self._experimentID,
                                                                                        cv.SESSION_ID)

                    wclause += ' & ( type == {0} ) '.format(event_type_id)

                    if filter_id is not None:
                        wclause += '& ( filter_id == {0} ) '.format(filter_id)

                    # start Conditions need to be added to where clause
                    if startConditions is not None:
                        wclause += '& ('
                        for conditionAttributeName, conditionAttributeComparitor in startConditions.items():
                            avComparison, value = conditionAttributeComparitor
                            value = self.getValuesForVariables(cv, value, cvNames)
                            wclause += ' ( {0} {1} {2} ) & '.format(conditionAttributeName, avComparison, value)
                        wclause = wclause[:-3]
                        wclause += ' ) '

                    # end Conditions need to be added to where clause
                    if endConditions is not None:
                        wclause += ' & ('
                        for conditionAttributeName, conditionAttributeComparitor in endConditions.items():
                            avComparison, value = conditionAttributeComparitor
                            value = self.getValuesForVariables(cv, value, cvNames)
                            wclause += ' ( {0} {1} {2} ) & '.format(conditionAttributeName, avComparison, value)
                        wclause = wclause[:-3]
                        wclause += ' ) '

                    for ename in event_attribute_names:
                        resultSetList[-1].append(getattr(deviceEventTable, read_where)(wclause, field=ename))
                    resultSetList[-1].append(wclause)
                    resultSetList[-1].append(cv)

                    eventAttributeResults = EventAttributeResults(*resultSetList[-1])
                    resultSetList[-1] = eventAttributeResults

                return resultSetList

            return None

    def getEventIterator(self, event_type):
        """
        **Docstr TBC.**

        Args:
            event_type

        Returns:
            (iterator): An iterator providing access to each matching event  as a numpy recarray.
        """
        return self.getEventTable(event_type).iterrows()

    def close(self):
        """Close the ExperimentDataAccessUtility and associated DataStore
        File."""
        global _hubFiles
        if self.hdfFile in _hubFiles:
            _hubFiles.remove(self.hdfFile)
        self.hdfFile.close()

        self.experimentCodes = None
        self.hdfFilePath = None
        self.hdfFileName = None
        self.mode = None
        self.hdfFile = None

    def __del__(self):
        try:
            self.close()
        except Exception:
            pass


class ExperimentDataAccessException(Exception):
    pass
